#include <lib94/lib94.hpp>
#include <functional>
#include <cassert>
#include <random>
#include <queue>
#include <set>

static std::mt19937_64 prng;
instruction core[CORE_SIZE];

void seed_prng(uint_fast64_t seed) {
  prng.seed(seed);
}

void clear_core(const instruction &background) {
  for (instruction &i : core)
    i = background;
}

void clear_core_random() {
  std::uniform_int_distribution<uint8_t> op(0, 15);
  std::uniform_int_distribution<uint8_t> mod(0, 6);
  std::uniform_int_distribution<uint8_t> modes(0, 7);
  std::uniform_int_distribution<number_t> number(0, CORE_SIZE - 1);

  for (instruction &i : core)
    i = {
      .op = (opcode)op(prng),
      .mod = (modifier)mod(prng),
      .amode = (mode)modes(prng),
      .bmode = (mode)modes(prng),
      .anumber = number(prng),
      .bnumber = number(prng)
    };
}

static std::set<number_t> modified_addresses;

struct warrior_info {
  const warrior *the_warrior;
  std::queue<number_t> processes;
};

static std::vector<warrior_info> warrior_infos;
std::queue<warrior_info *> alive_warriors;

bool init_round(const std::vector<const warrior *> &warriors) {
  modified_addresses.clear();
  warrior_infos.clear();
  alive_warriors = std::queue<warrior_info *>();

  std::uniform_int_distribution<number_t> number(0, CORE_SIZE - 1);
  std::vector<std::pair<number_t, number_t>> placements;

  for (const warrior *w : warriors) {
    unsigned tries = 0;

  new_place_at:
    if (tries > 1000)
      return false;
    ++tries;

    number_t place_at = number(prng);

    for (std::pair<number_t, number_t> &other : placements)
      //there has to be a better way
      for (number_t i = 0; i < w->instructions.size(); ++i)
        if (((place_at + i) % CORE_SIZE >= other.first && (place_at + i) % CORE_SIZE < other.first + other.second) ||
            ((place_at + i) % CORE_SIZE + CORE_SIZE >= other.first && (place_at + i) % CORE_SIZE + CORE_SIZE < other.first + other.second))
          goto new_place_at;

    placements.push_back(std::make_pair<>(place_at, w->instructions.size()));

    for (number_t i = 0; i < w->instructions.size(); ++i)
      core[(place_at + i) % CORE_SIZE] = w->instructions[i];

    warrior_infos.push_back({});
    warrior_info *wi = &warrior_infos.back();
    wi->the_warrior = w;
    wi->processes.push((place_at + w->org) % CORE_SIZE);
  }

  std::shuffle(warrior_infos.begin(), warrior_infos.end(), prng);

  for (warrior_info &wi : warrior_infos)
    alive_warriors.push(&wi);

  return true;
}

size_t alive_warrior_count() {
  return alive_warriors.size();
}

const warrior *get_next_warrior() {
  assert(alive_warriors.size() > 0);
  return alive_warriors.front()->the_warrior;
}

const std::queue<number_t> &get_processes(const warrior *for_warrior) {
  for (const warrior_info &wi : warrior_infos)
    if (wi.the_warrior == for_warrior)
      return wi.processes;
  assert(false);
}

number_t get_next_process(const warrior *for_warrior) {
  for (const warrior_info &wi : warrior_infos)
    if (wi.the_warrior == for_warrior) {
      assert(wi.processes.size() > 0);
      return wi.processes.front();
    }
  assert(false);
}

const std::set<number_t> &get_modified_addresses() {
  return modified_addresses;
}

void clear_modified_addresses() {
  modified_addresses.clear();
}

void add_modified_instruction(const instruction *instr) {
  modified_addresses.insert(instr - core);
}

const instruction &get_instruction(number_t address) {
  return core[address];
}

static const warrior *single_step_return_dead();

void single_step() {
  single_step_return_dead();
}

const warrior *run_until_warrior_death() {
  while (true) {
    const warrior *w = single_step_return_dead();
    if (w)
      return w;
  }
}

const warrior *run_until_winner() {
  while (true) {
    const warrior *w = single_step_return_dead();
    if (w && alive_warriors.size() == 1)
      for (const warrior_info &wi : warrior_infos)
        if (wi.processes.size() > 0)
          return wi.the_warrior;
  }
}

number_t program_counter;
static instruction instruction_register;

static number_t a_pointer, b_pointer;
instruction a_instruction, b_instruction;

instruction *a_instruction_writable;
instruction *b_instruction_writable;

static void evaluate_operand(mode m, number_t n, number_t &ptr, instruction &instr) {
  instruction *secondary;

  switch (m) {

    case IMMEDIATE:
      ptr = 0;
      instr = core[program_counter];
      break;

    case DIRECT:
      ptr = n;
      instr = core[(program_counter + n) % CORE_SIZE];
      break;

    case A_INDIRECT:
      secondary = core + (program_counter + n) % CORE_SIZE;
      ptr = (n + secondary->anumber) % CORE_SIZE;
      instr = core[(program_counter + ptr) % CORE_SIZE];
      break;

    case B_INDIRECT:
      secondary = core + (program_counter + n) % CORE_SIZE;
      ptr = (n + secondary->bnumber) % CORE_SIZE;
      instr = core[(program_counter + ptr) % CORE_SIZE];
      break;

    case A_DECREMENT:
      secondary = core + (program_counter + n) % CORE_SIZE;
      ptr = (n + --secondary->anumber) % CORE_SIZE;
      instr = core[(program_counter + ptr) % CORE_SIZE];
      add_modified_instruction(secondary);
      break;

    case B_DECREMENT:
      secondary = core + (program_counter + n) % CORE_SIZE;
      ptr = (n + --secondary->bnumber) % CORE_SIZE;
      instr = core[(program_counter + ptr) % CORE_SIZE];
      add_modified_instruction(secondary);
      break;

    case A_INCREMENT:
      secondary = core + (program_counter + n) % CORE_SIZE;
      ptr = (n + secondary->anumber) % CORE_SIZE;
      instr = core[(program_counter + ptr) % CORE_SIZE];
      ++secondary->anumber;
      add_modified_instruction(secondary);
      break;

    case B_INCREMENT:
      secondary = core + (program_counter + n) % CORE_SIZE;
      ptr = (n + secondary->bnumber) % CORE_SIZE;
      instr = core[(program_counter + ptr) % CORE_SIZE];
      ++secondary->bnumber;
      add_modified_instruction(secondary);
      break;

  }
}

template<opcode op, modifier mod>
bool execute();

#define EXECUTOR_DECL_LINE(mod) \
  extern template bool execute<DAT, mod>(); extern template bool execute<MOV, mod>(); \
  extern template bool execute<ADD, mod>(); extern template bool execute<SUB, mod>(); \
  extern template bool execute<MUL, mod>(); extern template bool execute<DIV, mod>(); \
  extern template bool execute<MOD, mod>(); extern template bool execute<JMP, mod>(); \
  extern template bool execute<JMZ, mod>(); extern template bool execute<JMN, mod>(); \
  extern template bool execute<DJN, mod>(); extern template bool execute<SEQ, mod>(); \
  extern template bool execute<SNE, mod>(); extern template bool execute<SLT, mod>(); \
  extern template bool execute<SPL, mod>(); extern template bool execute<NOP, mod>();

EXECUTOR_DECL_LINE(A)
EXECUTOR_DECL_LINE(B)
EXECUTOR_DECL_LINE(AB)
EXECUTOR_DECL_LINE(BA)
EXECUTOR_DECL_LINE(F)
EXECUTOR_DECL_LINE(X)
EXECUTOR_DECL_LINE(I)

#define EXECUTOR_REF_LINE(mod) \
  &execute<DAT, mod>, &execute<MOV, mod>, \
  &execute<ADD, mod>, &execute<SUB, mod>, \
  &execute<MUL, mod>, &execute<DIV, mod>, \
  &execute<MOD, mod>, &execute<JMP, mod>, \
  &execute<JMZ, mod>, &execute<JMN, mod>, \
  &execute<DJN, mod>, &execute<SEQ, mod>, \
  &execute<SNE, mod>, &execute<SLT, mod>, \
  &execute<SPL, mod>, &execute<NOP, mod>

static const std::function<bool ()> executors[] = {
  EXECUTOR_REF_LINE(A),  EXECUTOR_REF_LINE(B),
  EXECUTOR_REF_LINE(AB), EXECUTOR_REF_LINE(BA),
  EXECUTOR_REF_LINE(F),  EXECUTOR_REF_LINE(X),
  EXECUTOR_REF_LINE(I)
};

static warrior_info *this_warrior;

void enqueue_process(number_t pc) {
  this_warrior->processes.push(pc);
}

static const warrior *single_step_return_dead() {
  this_warrior = alive_warriors.front();
  alive_warriors.pop();

  program_counter = this_warrior->processes.front();
  this_warrior->processes.pop();

  instruction_register = core[program_counter];

  evaluate_operand(instruction_register.amode, instruction_register.anumber, a_pointer, a_instruction);
  evaluate_operand(instruction_register.bmode, instruction_register.bnumber, b_pointer, b_instruction);

  a_instruction_writable = core + (program_counter + a_pointer) % CORE_SIZE;
  b_instruction_writable = core + (program_counter + b_pointer) % CORE_SIZE;

  bool enqueue_process = executors[instruction_register.op + instruction_register.mod * 16]();

  if (enqueue_process)
    this_warrior->processes.push((program_counter + 1) % CORE_SIZE);

  else if (this_warrior->processes.size() == 0)
    return this_warrior->the_warrior;

  alive_warriors.push(this_warrior);
  return 0;
}
